import pytest
import requests_mock
from openfeature.evaluation_context import EvaluationContext

from confidence.confidence import Confidence
from confidence.openfeature_provider import ConfidenceOpenFeatureProvider
from tests.test_provider import SUCCESSFUL_FLAG_RESOLVE


@pytest.mark.parametrize("apply_on_resolve", ((True, False)))
def test_apply_configurable(apply_on_resolve):
    ctx = EvaluationContext(targeting_key="meh")
    apply_provider = ConfidenceOpenFeatureProvider(
        Confidence(client_secret="test", apply_on_resolve=apply_on_resolve)
    )
    with requests_mock.Mocker() as mock:
        mock.post(
            "https://resolver.confidence.dev/v1/flags:resolve",
            json=SUCCESSFUL_FLAG_RESOLVE,
        )

        apply_provider.resolve_string_details(
            flag_key="flags/python-flag-1.string-key",
            default_value="yellow",
            evaluation_context=ctx,
        )

        last_request = mock.request_history[-1]
        assert last_request.json()["apply"] == apply_on_resolve



import asyncio
import base64
import dataclasses
from datetime import datetime
from enum import Enum
import json
import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
)

import requests
import httpx
from typing_extensions import TypeGuard

from confidence import __version__
from confidence.errors import (
    FlagNotFoundError,
    GeneralError,
    ParseError,
    TypeMismatchError,
    TimeoutError,
)
from .flag_types import FlagResolutionDetails, Reason, ErrorCode
from .names import FlagName, VariantName

EU_RESOLVE_API_ENDPOINT = "https://resolver.eu.confidence.dev"
US_RESOLVE_API_ENDPOINT = "https://resolver.us.confidence.dev"
GLOBAL_RESOLVE_API_ENDPOINT = "https://resolver.confidence.dev"

# Default timeout in milliseconds (10 seconds)
DEFAULT_TIMEOUT_MS = 10000

Primitive = Union[str, int, float, bool, None]
FieldType = Union[Primitive, List[Primitive], List["Object"], "Object"]
Object = Dict[str, FieldType]


def is_primitive(field_type: Type[Any]) -> TypeGuard[Type[Primitive]]:
    return field_type in get_args(Primitive)


def primitive_matches(value: FieldType, value_type: Type[Primitive]) -> bool:
    return (
        value_type is None
        or (value_type is int and isinstance(value, int))
        or (value_type is float and isinstance(value, float))
        or (value_type is str and isinstance(value, str))
        or (value_type is bool and isinstance(value, bool))
    )


class Region(Enum):
    def endpoint(self) -> str:
        return self.value

    EU = EU_RESOLVE_API_ENDPOINT
    US = US_RESOLVE_API_ENDPOINT
    GLOBAL = GLOBAL_RESOLVE_API_ENDPOINT


@dataclasses.dataclass
class ResolveResult(object):
    value: Optional[Object]
    variant: Optional[str]
    token: str


class Confidence:
    context: Dict[str, FieldType] = {}

    def put_context(self, key: str, value: FieldType) -> None:
        self.context[key] = value

    def with_context(self, context: Dict[str, FieldType]) -> "Confidence":
        new_confidence = Confidence(
            self._client_secret,
            self._region,
            self._apply_on_resolve,
            self._custom_resolve_base_url,
            timeout_ms=self._timeout_ms,
            logger=self.logger,
            async_client=self.async_client,
        )
        new_confidence.context = {**self.context, **context}
        return new_confidence

    def __init__(
        self,
        client_secret: str,
        region: Region = Region.GLOBAL,
        apply_on_resolve: bool = True,
        custom_resolve_base_url: Optional[str] = None,
        timeout_ms: Optional[int] = DEFAULT_TIMEOUT_MS,
        logger: logging.Logger = logging.getLogger("confidence_logger"),
        async_client: httpx.AsyncClient = httpx.AsyncClient(),
    ):
        self._client_secret = client_secret
        self._region = region
        self._api_endpoint = region.endpoint()
        self._apply_on_resolve = apply_on_resolve
        self._timeout_ms = timeout_ms
        self.logger = logger
        self.async_client = async_client
        self._setup_logger(logger)
        self._custom_resolve_base_url = custom_resolve_base_url

    def resolve_boolean_details(
        self, flag_key: str, default_value: bool
    ) -> FlagResolutionDetails[bool]:
        return self._evaluate(flag_key, bool, default_value, self.context)

    async def resolve_boolean_details_async(
        self, flag_key: str, default_value: bool
    ) -> FlagResolutionDetails[bool]:
        return await self._evaluate_async(flag_key, bool, default_value, self.context)

    def resolve_float_details(
        self, flag_key: str, default_value: float
    ) -> FlagResolutionDetails[float]:
        return self._evaluate(flag_key, float, default_value, self.context)

    async def resolve_float_details_async(
        self, flag_key: str, default_value: float
    ) -> FlagResolutionDetails[float]:
        return await self._evaluate_async(flag_key, float, default_value, self.context)

    def resolve_integer_details(
        self, flag_key: str, default_value: int
    ) -> FlagResolutionDetails[int]:
        return self._evaluate(flag_key, int, default_value, self.context)

    async def resolve_integer_details_async(
        self, flag_key: str, default_value: int
    ) -> FlagResolutionDetails[int]:
        return await self._evaluate_async(flag_key, int, default_value, self.context)

    def resolve_string_details(
        self, flag_key: str, default_value: str
    ) -> FlagResolutionDetails[str]:
        return self._evaluate(flag_key, str, default_value, self.context)

    async def resolve_string_details_async(
        self, flag_key: str, default_value: str
    ) -> FlagResolutionDetails[str]:
        return await self._evaluate_async(flag_key, str, default_value, self.context)

    def resolve_object_details(
        self, flag_key: str, default_value: Union[Object, List[Primitive]]
    ) -> FlagResolutionDetails[Union[Object, List[Primitive]]]:
        return self._evaluate(flag_key, Object, default_value, self.context)

    async def resolve_object_details_async(
        self, flag_key: str, default_value: Union[Object, List[Primitive]]
    ) -> FlagResolutionDetails[Union[Object, List[Primitive]]]:
        return await self._evaluate_async(flag_key, Object, default_value, self.context)

    #
    # --- internals
    #

    def _setup_logger(self, logger: logging.Logger) -> None:
        if logger is not None:
            logger.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
            if not logger.hasHandlers():
                ch = logging.StreamHandler()
                ch.setFormatter(formatter)
                logger.addHandler(ch)

    def _logResolveTester(self, flag_id: str, context: Dict[str, FieldType]) -> None:
        json_payload = json.dumps(
            {
                "flag": f"flags/{flag_id}",
                "context": context,
                "clientKey": self._client_secret,
            }
        )
        base64_payload = base64.b64encode(json_payload.encode("utf-8")).decode("utf-8")
        self.logger.debug(
            f"Check your flag evaluation for '{flag_id}' by copy-pasting the payload to the Resolve tester: {base64_payload}"  # noqa: E501
        )

    def _handle_evaluation_result(
        self,
        result: ResolveResult,
        flag_id: str,
        flag_key: str,
        value_type: Type[FieldType],
        default_value: FieldType,
        value_path: Optional[str],
        context: Dict[str, FieldType],
    ) -> FlagResolutionDetails[Any]:
        self._logResolveTester(flag_id, context)

        if result.variant is None or len(str(result.value)) == 0:
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                flag_metadata={"flag_key": flag_key},
            )

        variant_name = VariantName.parse(result.variant)

        value = self._select(result, value_path, value_type, self.logger)
        if value is None:
            self.logger.debug(
                f"Flag {flag_key} resolved to None. Returning default value."
            )
            value = default_value

        return FlagResolutionDetails(
            value=value,
            variant=variant_name.variant,
            reason=Reason.TARGETING_MATCH,
            flag_metadata={"flag_key": flag_key},
        )

    def _evaluate(
        self,
        flag_key: str,
        value_type: Type[FieldType],
        default_value: FieldType,
        context: Dict[str, FieldType],
    ) -> FlagResolutionDetails[Any]:
        if "." in flag_key:
            flag_id, value_path = flag_key.split(".", 1)
        else:
            flag_id = flag_key
            value_path = None
        try:
            result = self._resolve(FlagName(flag_id), context)
            return self._handle_evaluation_result(
                result,
                flag_id,
                flag_key,
                value_type,
                default_value,
                value_path,
                context,
            )
        except FlagNotFoundError:
            self.logger.info(f"Flag {flag_key} not found")
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                error_code=ErrorCode.FLAG_NOT_FOUND,
                error_message=f"Flag {flag_key} not found",
                flag_metadata={"flag_key": flag_key},
            )
        except TimeoutError as e:
            self.logger.warning(
                f"Request timed out after {self._timeout_ms} ms"
                f" when resolving flag {flag_key}"
            )
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                error_code=ErrorCode.TIMEOUT,
                error_message=str(e),
                flag_metadata={"flag_key": flag_key},
            )
        except Exception as e:
            self.logger.error(f"Error resolving flag {flag_key}: {str(e)}")
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.ERROR,
                error_code=ErrorCode.GENERAL,
                error_message=str(e),
                flag_metadata={"flag_key": flag_key},
            )

    async def _evaluate_async(
        self,
        flag_key: str,
        value_type: Type[FieldType],
        default_value: FieldType,
        context: Dict[str, FieldType],
    ) -> FlagResolutionDetails[Any]:
        if "." in flag_key:
            flag_id, value_path = flag_key.split(".", 1)
        else:
            flag_id = flag_key
            value_path = None
        try:
            result = await self._resolve_async(FlagName(flag_id), context)
            return self._handle_evaluation_result(
                result,
                flag_id,
                flag_key,
                value_type,
                default_value,
                value_path,
                context,
            )
        except FlagNotFoundError:
            self.logger.info(f"Flag {flag_key} not found")
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                error_code=ErrorCode.FLAG_NOT_FOUND,
                error_message=f"Flag {flag_key} not found",
                flag_metadata={"flag_key": flag_key},
            )
        except TimeoutError as e:
            self.logger.warning(
                f"Request timed out after {self._timeout_ms} ms"
                f" when resolving flag {flag_key}"
            )
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                error_code=ErrorCode.TIMEOUT,
                error_message=str(e),
                flag_metadata={"flag_key": flag_key},
            )
        except Exception as e:
            self.logger.error(f"Error resolving flag {flag_key}: {str(e)}")
            return FlagResolutionDetails(
                value=default_value,
                reason=Reason.DEFAULT,
                error_code=ErrorCode.GENERAL,
                error_message=str(e),
                flag_metadata={"flag_key": flag_key},
            )

    # type-arg: ignore
    def track(self, event_name: str, data: Dict[str, FieldType]) -> None:
        self._send_event_internal(event_name, data)

    def track_async(self, event_name: str, data: Dict[str, FieldType]) -> None:
        asyncio.create_task(self._send_event(event_name, data))

    async def _send_event(self, event_name: str, data: Dict[str, FieldType]) -> None:
        self._send_event_internal(event_name, data)

    def _send_event_internal(self, event_name: str, data: Dict[str, FieldType]) -> None:
        current_time = datetime.utcnow().isoformat() + "Z"
        request_body = {
            "clientSecret": self._client_secret,
            "sendTime": current_time,
            "events": [
                {
                    "eventDefinition": f"eventDefinitions/{event_name}",
                    "payload": {"context": {**self.context}, **data},
                    "eventTime": current_time,
                }
            ],
            "sdk": {"id": "SDK_ID_PYTHON_CONFIDENCE", "version": __version__},
        }

        event_url = "https://events.confidence.dev/v1/events:publish"
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        timeout_sec = None if self._timeout_ms is None else self._timeout_ms / 1000.0
        try:
            response = requests.post(
                event_url, json=request_body, headers=headers, timeout=timeout_sec
            )
            if response.status_code == 200:
                json = response.json()

                json_errors = json.get("errors")
                if json_errors:
                    self.logger.warning("events emitted with errors:")
                    for error in json_errors:
                        self.logger.warning(error)
            else:
                self.logger.warning(
                    f"Track event {event_name} failed with status code"
                    + f" {response.status_code} and reason: {response.reason}"
                )
        except requests.exceptions.RequestException as e:
            self.logger.warning(f"Failed to track event {event_name}: {str(e)}")

    def _handle_resolve_response(
        self, response: requests.Response, flag_name: FlagName
    ) -> ResolveResult:
        if response.status_code == 404:
            self.logger.error(f"Flag {flag_name} not found")
            raise FlagNotFoundError()

        response.raise_for_status()

        response_body = response.json()

        resolved_flags = response_body["resolvedFlags"]
        token = response_body["resolveToken"]

        if len(resolved_flags) == 0:
            raise FlagNotFoundError()

        resolved_flag = resolved_flags[0]
        variant = resolved_flag.get("variant")
        return ResolveResult(
            resolved_flag.get("value"), None if variant == "" else variant, token
        )

    def _resolve(
        self, flag_name: FlagName, context: Dict[str, FieldType]
    ) -> ResolveResult:
        request_body = {
            "clientSecret": self._client_secret,
            "evaluationContext": context,
            "apply": self._apply_on_resolve,
            "flags": [str(flag_name)],
            "sdk": {"id": "SDK_ID_PYTHON_CONFIDENCE", "version": __version__},
        }
        base_url = self._api_endpoint
        if self._custom_resolve_base_url is not None:
            base_url = self._custom_resolve_base_url

        resolve_url = f"{base_url}/v1/flags:resolve"
        timeout_sec = None if self._timeout_ms is None else self._timeout_ms / 1000.0
        try:
            response = requests.post(
                resolve_url, json=request_body, timeout=timeout_sec
            )
            return self._handle_resolve_response(response, flag_name)
        except requests.exceptions.Timeout:
            self.logger.warning(
                f"Request timed out after {timeout_sec}s"
                f" when resolving flag {flag_name}"
            )
            raise TimeoutError()
        except requests.exceptions.RequestException as e:
            self.logger.warning(f"Error resolving flag {flag_name}: {str(e)}")
            raise GeneralError(str(e))

    async def _resolve_async(
        self, flag_name: FlagName, context: Dict[str, FieldType]
    ) -> ResolveResult:
        request_body = {
            "clientSecret": self._client_secret,
            "evaluationContext": context,
            "apply": self._apply_on_resolve,
            "flags": [str(flag_name)],
            "sdk": {"id": "SDK_ID_PYTHON_CONFIDENCE", "version": __version__},
        }
        base_url = self._api_endpoint
        if self._custom_resolve_base_url is not None:
            base_url = self._custom_resolve_base_url

        resolve_url = f"{base_url}/v1/flags:resolve"
        timeout_sec = None if self._timeout_ms is None else self._timeout_ms / 1000.0
        try:
            response = await self.async_client.post(
                resolve_url, json=request_body, timeout=timeout_sec
            )
            return self._handle_resolve_response(response, flag_name)
        except httpx.TimeoutException:
            self.logger.warning(
                f"Request timed out after {timeout_sec}s"
                f" when resolving flag {flag_name}"
            )
            raise TimeoutError()
        except httpx.HTTPError as e:
            self.logger.warning(f"Error resolving flag {flag_name}: {str(e)}")
            raise GeneralError(str(e))

    @staticmethod
    def _select(
        result: ResolveResult,
        value_path: Optional[str],
        value_type: Type[FieldType],
        logger: logging.Logger,
    ) -> FieldType:
        value: FieldType = result.value

        if value_path is not None:
            keys = value_path.split(".")
            for key in keys:
                if not isinstance(value, dict):
                    logger.debug(f"Value {value} is not a dict. Returning None.")
                    raise ParseError()

                if key not in value:
                    logger.debug(
                        f"Key {key} not found in value {value}. Returning None."
                    )
                    raise ParseError()

                value = value.get(key)

        # skip type checking if the value was not specified
        if value is None:
            return None

        if not Confidence.type_matches(value, value_type):
            logger.debug(
                f"Type of value {value} did not match expected type {value_type}."
            )
            raise TypeMismatchError("type of value did not match excepted type")

        return value

    @staticmethod
    def type_matches(value: FieldType, value_type: Type[FieldType]) -> bool:
        origin = get_origin(value_type)

        if is_primitive(value_type):
            return primitive_matches(value, value_type)
        elif origin is list:
            return isinstance(value, list)
        elif origin is dict:
            return isinstance(value, dict)

        return False



import requests_mock
import unittest
from unittest.mock import patch, AsyncMock
import json
import httpx
from requests.exceptions import Timeout as RequestsTimeout


import confidence.confidence
from confidence.confidence import Confidence, DEFAULT_TIMEOUT_MS
from confidence.errors import ErrorCode
from confidence.flag_types import Reason


class TestConfidence(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.confidence = Confidence(client_secret="test")

    def test_resolve_string_with_dot_notation(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.string-key"
            )
            self.assertEqual(result.value, "outer-string")
            self.assertEqual(result.variant, "enabled")
            self.assertEqual(result.reason, Reason.TARGETING_MATCH)

    def test_resolve_failed(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=NO_MATCH_STRING_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_string_details(
                flag_key="some-flag-that-doesnt-exist",
                default_value="yellow",
            )

            self.assertEqual(result.reason, Reason.DEFAULT)
            self.assertEqual(
                result.flag_metadata["flag_key"], "some-flag-that-doesnt-exist"
            )
            self.assertEqual(result.value, "yellow")
            self.assertIsNone(result.variant)
            self.assertEqual(result.error_code, ErrorCode.FLAG_NOT_FOUND)
            self.assertEqual(
                result.error_message, "Flag some-flag-that-doesnt-exist not found"
            )

    def test_resolve_string_with_dot_notation_request_payload(self):
        with requests_mock.Mocker() as mock:
            confidence.confidence.__version__ = "v0.0.0"
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            self.confidence.with_context(
                {
                    "targeting_key": "boop",
                    "user": {"country": "US"},
                    "connection": "wifi",
                }
            ).resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            last_request = mock.request_history[-1]
            self.assertEqual(last_request.method, "POST")
            self.assertEqual(
                last_request.url, "https://resolver.confidence.dev/v1/flags:resolve"
            )
            self.assertEqual(
                last_request.json(),
                EXPECTED_REQUEST_PAYLOAD,
            )

    def test_resolve_successful_custom_url(self):
        self.confidence = Confidence(
            client_secret="test", custom_resolve_base_url="https://custom_url"
        )

        with requests_mock.Mocker() as mock:
            mock.post(
                "https://custom_url/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_object_details(
                flag_key="python-flag-1.struct-key",
                default_value={"key": "value"},
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.struct-key"
            )
            self.assertEqual(result.value, {"string-key": "inner-string"})

    def test_resolve_response_object_details(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_object_details(
                flag_key="python-flag-1",
                default_value={"key": "value"},
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1")
            self.assertEqual(
                result.value,
                {
                    "double-key": 42.42,
                    "enabled": True,
                    "int-key": 42,
                    "string-key": "outer-string",
                    "struct-key": {"string-key": "inner-string"},
                },
            )

    async def test_resolve_response_object_details_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        with patch("httpx.AsyncClient.post", return_value=mock_response):
            result = await self.confidence.resolve_object_details_async(
                flag_key="python-flag-1", default_value={"key": "value"}
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1")
            self.assertEqual(
                result.value,
                {
                    "double-key": 42.42,
                    "enabled": True,
                    "int-key": 42,
                    "string-key": "outer-string",
                    "struct-key": {"string-key": "inner-string"},
                },
            )

    def test_resolve_struct_details(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_object_details(
                flag_key="python-flag-1.struct-key",
                default_value={"key": "value"},
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.struct-key"
            )
            self.assertEqual(result.value, {"string-key": "inner-string"})

    async def test_resolve_struct_details_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        with patch("httpx.AsyncClient.post", return_value=mock_response):
            result = await self.confidence.resolve_object_details_async(
                flag_key="python-flag-1.struct-key", default_value={"key": "value"}
            )

            assert result.reason == Reason.TARGETING_MATCH
            assert result.flag_metadata["flag_key"] == "python-flag-1.struct-key"
            assert result.value == {"string-key": "inner-string"}

    def test_resolve_integer_details(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_integer_details(
                flag_key="python-flag-1.int-key",
                default_value=-1,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.int-key")
            self.assertEqual(result.value, 42)

    async def test_resolve_integer_details_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        with patch("httpx.AsyncClient.post", return_value=mock_response):
            result = await self.confidence.resolve_integer_details_async(
                flag_key="python-flag-1.int-key", default_value=-1
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.int-key")
            self.assertEqual(result.value, 42)

    def test_resolve_boolean_details(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_boolean_details(
                flag_key="python-flag-1.enabled",
                default_value=False,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.enabled")
            self.assertEqual(result.value, True)

    async def test_resolve_boolean_details_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        with patch("httpx.AsyncClient.post", return_value=mock_response):
            result = await self.confidence.resolve_boolean_details_async(
                flag_key="python-flag-1.enabled", default_value=False
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.enabled")
            self.assertEqual(result.value, True)

    def test_resolve_float_details(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.resolve_float_details(
                flag_key="python-flag-1.double-key",
                default_value=0.01,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.double-key"
            )
            self.assertEqual(result.value, 42.42)

    async def test_resolve_float_details_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        with patch("httpx.AsyncClient.post", return_value=mock_response):
            result = await self.confidence.resolve_float_details_async(
                flag_key="python-flag-1.double-key", default_value=0.01
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.double-key"
            )
            self.assertEqual(result.value, 42.42)

    def test_resolve_without_targeting_key(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.confidence.with_context(
                {"connection": "wifi"}
            ).resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="brown",
            )
            self.assertIsNotNone(result.value)

            last_request = mock.request_history[-1]
            self.assertIsNone(
                last_request.json()["evaluationContext"].get("targeting_key"),
            )
            self.assertEqual(
                last_request.json()["evaluationContext"].get("connection"), "wifi"
            )

    def test_no_segment_match(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=NO_SEGMENT_MATCH_STRING_FLAG_RESOLVE,
            )
            result = self.confidence.with_context(
                {"connection": "wifi"}
            ).resolve_string_details(
                flag_key="test-flag.color",
                default_value="brown",
            )
            self.assertEqual(result.value, "brown")
            self.assertEqual(result.reason, Reason.DEFAULT)

    def test_resolve_with_timeout(self):
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )

            with patch("requests.post") as mock_post:
                mock_post.return_value.status_code = 200
                mock_post.return_value.json.return_value = SUCCESSFUL_FLAG_RESOLVE

                confidence_with_timeout = Confidence(
                    client_secret="test", timeout_ms=5500
                )
                confidence_with_timeout.resolve_string_details(
                    flag_key="python-flag-1.string-key",
                    default_value="yellow",
                )

                # Verify that timeout was passed to requests.post
                mock_post.assert_called_once()
                _, kwargs = mock_post.call_args
                self.assertEqual(kwargs["timeout"], 5.5)

    async def test_resolve_with_timeout_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        # Create an AsyncMock that returns an awaitable that resolves to mock_response
        mock_post = AsyncMock()
        mock_post.return_value = mock_response

        with patch("httpx.AsyncClient.post", mock_post):
            confidence_with_timeout = Confidence(client_secret="test", timeout_ms=3500)
            await confidence_with_timeout.resolve_string_details_async(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            # Verify that timeout was passed to httpx.AsyncClient.post
            mock_post.assert_called_once()
            _, kwargs = mock_post.call_args
            self.assertEqual(kwargs["timeout"], 3.5)

    def test_resolve_with_default_timeout(self):
        """Test that the default timeout is used when timeout_ms is not provided."""
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )

            with patch("requests.post") as mock_post:
                mock_post.return_value.status_code = 200
                mock_post.return_value.json.return_value = SUCCESSFUL_FLAG_RESOLVE

                # Create client without specifying timeout_ms
                confidence_default_timeout = Confidence(client_secret="test")
                confidence_default_timeout.resolve_string_details(
                    flag_key="python-flag-1.string-key",
                    default_value="yellow",
                )

                # Verify that default timeout (10 seconds) was passed to requests.post
                mock_post.assert_called_once()
                _, kwargs = mock_post.call_args
                self.assertEqual(kwargs["timeout"], DEFAULT_TIMEOUT_MS / 1000.0)

    async def test_resolve_with_default_timeout_async(self):
        mock_response = httpx.Response(
            status_code=200,
            json=SUCCESSFUL_FLAG_RESOLVE,
            request=httpx.Request(
                "POST", "https://resolver.confidence.dev/v1/flags:resolve"
            ),
        )

        # Create an AsyncMock that returns an awaitable that resolves to mock_response
        mock_post = AsyncMock()
        mock_post.return_value = mock_response

        with patch("httpx.AsyncClient.post", mock_post):
            # Create client without specifying timeout_ms
            confidence_default_timeout = Confidence(client_secret="test")
            await confidence_default_timeout.resolve_string_details_async(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            # Verify that default timeout (10 seconds) was passed to
            # httpx.AsyncClient.post
            mock_post.assert_called_once()
            _, kwargs = mock_post.call_args
            self.assertEqual(kwargs["timeout"], DEFAULT_TIMEOUT_MS / 1000.0)

    def test_handle_actual_timeout(self):
        with patch("requests.post") as mock_post:
            # Simulate a timeout by raising the Timeout exception
            mock_post.side_effect = RequestsTimeout("Connection timed out")

            confidence_with_timeout = Confidence(client_secret="test", timeout_ms=100)

            # The operation should NOT raise an exception, but return default value
            result = confidence_with_timeout.resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            # Verify that default value was returned
            self.assertEqual(result.value, "yellow")
            self.assertEqual(result.reason, Reason.DEFAULT)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.string-key"
            )
            self.assertIsNone(result.variant)
            self.assertEqual(result.error_code, ErrorCode.TIMEOUT)

    async def test_handle_actual_timeout_async(self):
        # Create an AsyncMock that raises a timeout exception
        mock_post = AsyncMock()
        mock_post.side_effect = httpx.TimeoutException("Connection timed out")

        with patch("httpx.AsyncClient.post", mock_post):
            confidence_with_timeout = Confidence(client_secret="test", timeout_ms=100)

            # The operation should NOT raise an exception, but return default value
            result = await confidence_with_timeout.resolve_string_details_async(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
            )

            # Verify that default value was returned
            self.assertEqual(result.value, "yellow")
            self.assertEqual(result.reason, Reason.DEFAULT)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.string-key"
            )
            self.assertIsNone(result.variant)
            self.assertEqual(result.error_code, ErrorCode.TIMEOUT)

    if __name__ == "__main__":
        unittest.main()


NO_MATCH_STRING_FLAG_RESOLVE = json.loads(
    """{"resolvedFlags": [], "resolveToken": ""}"""
)

NO_SEGMENT_MATCH_STRING_FLAG_RESOLVE = json.loads(
    """{
  "resolveToken": "",
  "resolvedFlags": [
    {
      "flag": "flags/test-flag-2",
      "reason": "RESOLVE_REASON_NO_SEGMENT_MATCH",
      "variant": ""
    }
  ]
}"""
)

EXPECTED_REQUEST_PAYLOAD = json.loads(
    """{
  "clientSecret": "test",
  "evaluationContext": {
    "targeting_key": "boop",
    "user": { "country": "US" },
    "connection": "wifi"
  },
  "apply": true,
  "flags": ["flags/python-flag-1"],
  "sdk": {
    "id": "SDK_ID_PYTHON_CONFIDENCE",
    "version": "v0.0.0"
   }
}"""
)

SUCCESSFUL_FLAG_RESOLVE = json.loads(
    """{
     "resolvedFlags": [
    {
      "flag": "flags/python-flag-1",
      "variant": "flags/python-flag-1/variants/enabled",
      "value": {
        "struct-key": {
          "string-key": "inner-string"
        },
        "string-key": "outer-string",
        "double-key": 42.42,
        "int-key": 42,
        "enabled": true
      },
      "flagSchema": {
        "schema": {
          "struct-key": {
            "structSchema": {
              "schema": {
                "string-key": {
                  "stringSchema": {}
                }
              }
            }
          },
          "string-key": {
            "stringSchema": {}
          },
          "double-key": {
            "doubleSchema": {}
          },
          "int-key": {
            "intSchema": {}
          },
          "enabled": {
            "boolSchema": {}
          }
        }
      },
      "reason": "RESOLVE_REASON_MATCH"
    }
  ],
  "resolveToken": ""
    }"""
)



import dataclasses
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Type,
    Union,
    get_args,
)

import openfeature.exception as open_feature_exception

import openfeature.exception
from openfeature.api import EvaluationContext
from openfeature.api import Hook
from openfeature.flag_evaluation import FlagResolutionDetails
from openfeature.provider.metadata import Metadata
from openfeature.provider.provider import AbstractProvider
from typing_extensions import TypeGuard

import confidence.confidence
from confidence.errors import ErrorCode

EU_RESOLVE_API_ENDPOINT = "https://resolver.eu.confidence.dev/v1"
US_RESOLVE_API_ENDPOINT = "https://resolver.us.confidence.dev/v1"
GLOBAL_RESOLVE_API_ENDPOINT = "https://resolver.confidence.dev/v1"


Primitive = Union[str, int, float, bool, None]
FieldType = Union[Primitive, List[Primitive], List["Object"], "Object"]
Object = Dict[str, FieldType]


def is_primitive(field_type: Type[Any]) -> TypeGuard[Type[Primitive]]:
    return field_type in get_args(Primitive)


def primitive_matches(value: FieldType, value_type: Type[Primitive]) -> bool:
    return (
        value_type is None
        or (value_type is int and isinstance(value, int))
        or (value_type is float and isinstance(value, float))
        or (value_type is str and isinstance(value, str))
        or (value_type is bool and isinstance(value, bool))
    )


class Region(Enum):
    def endpoint(self) -> str:
        return self.value

    EU = EU_RESOLVE_API_ENDPOINT
    US = US_RESOLVE_API_ENDPOINT
    GLOBAL = GLOBAL_RESOLVE_API_ENDPOINT


@dataclasses.dataclass
class ResolveResult(object):
    value: Optional[Object]
    variant: Optional[str]
    token: str


def _to_openfeature_error_code(
    error_code: Optional[ErrorCode],
) -> Optional[open_feature_exception.ErrorCode]:
    """
    Convert a confidence error code to an openfeature error code
    :param error_code:
    :return:
    """
    if error_code is None:
        return None
    if error_code is ErrorCode.FLAG_NOT_FOUND:
        return openfeature.exception.ErrorCode.FLAG_NOT_FOUND
    if error_code is ErrorCode.TYPE_MISMATCH:
        return openfeature.exception.ErrorCode.TYPE_MISMATCH
    if error_code is ErrorCode.TARGETING_KEY_MISSING:
        return openfeature.exception.ErrorCode.TARGETING_KEY_MISSING
    if error_code is ErrorCode.INVALID_CONTEXT:
        return openfeature.exception.ErrorCode.INVALID_CONTEXT
    if error_code is ErrorCode.GENERAL:
        return openfeature.exception.ErrorCode.GENERAL
    if error_code is ErrorCode.PARSE_ERROR:
        return openfeature.exception.ErrorCode.PARSE_ERROR
    if error_code is ErrorCode.TIMEOUT:
        return openfeature.exception.ErrorCode.GENERAL
    if error_code is ErrorCode.NOT_READY:
        return openfeature.exception.ErrorCode.PROVIDER_NOT_READY


class ConfidenceOpenFeatureProvider(AbstractProvider):
    def __init__(self, confidence_sdk: confidence.confidence.Confidence):
        self.confidence_sdk = confidence_sdk

    #
    # --- Provider API ---
    #

    def get_metadata(self) -> Metadata:
        return Metadata("Confidence")

    def get_provider_hooks(self) -> List[Hook]:
        return []

    def resolve_boolean_details(
        self,
        flag_key: str,
        default_value: bool,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[bool]:
        details = self._confidence_with_context(
            evaluation_context
        ).resolve_boolean_details(flag_key, default_value)
        return FlagResolutionDetails[bool](
            value=details.value,
            variant=details.variant,
            reason=details.reason,
            error_code=_to_openfeature_error_code(details.error_code),
            error_message=details.error_message,
            flag_metadata=details.flag_metadata,
        )

    def resolve_float_details(
        self,
        flag_key: str,
        default_value: float,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[float]:
        details = self._confidence_with_context(
            evaluation_context
        ).resolve_float_details(flag_key, default_value)
        return FlagResolutionDetails[float](
            value=details.value,
            variant=details.variant,
            reason=details.reason,
            error_code=_to_openfeature_error_code(details.error_code),
            error_message=details.error_message,
            flag_metadata=details.flag_metadata,
        )

    def resolve_integer_details(
        self,
        flag_key: str,
        default_value: int,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[int]:
        details = self._confidence_with_context(
            evaluation_context
        ).resolve_integer_details(flag_key, default_value)
        return FlagResolutionDetails[int](
            value=details.value,
            variant=details.variant,
            reason=details.reason,
            error_code=_to_openfeature_error_code(details.error_code),
            error_message=details.error_message,
            flag_metadata=details.flag_metadata,
        )

    def resolve_string_details(
        self,
        flag_key: str,
        default_value: str,
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[str]:
        details = self._confidence_with_context(
            evaluation_context
        ).resolve_string_details(flag_key, default_value)
        return FlagResolutionDetails[str](
            value=details.value,
            variant=details.variant,
            reason=details.reason,
            error_code=_to_openfeature_error_code(details.error_code),
            error_message=details.error_message,
            flag_metadata=details.flag_metadata,
        )

    def resolve_object_details(
        self,
        flag_key: str,
        default_value: Union[Object, List[Primitive]],
        evaluation_context: Optional[EvaluationContext] = None,
    ) -> FlagResolutionDetails[Union[Object, List[Primitive]]]:
        details = self._confidence_with_context(
            evaluation_context
        ).resolve_object_details(flag_key, default_value)
        return FlagResolutionDetails[Union[Object, List[Primitive]]](
            value=details.value,
            variant=details.variant,
            reason=details.reason,
            error_code=_to_openfeature_error_code(details.error_code),
            error_message=details.error_message,
            flag_metadata=details.flag_metadata,
        )

    def _confidence_with_context(
        self, evaluation_context: Optional[EvaluationContext]
    ) -> confidence.confidence.Confidence:
        eval_context: Dict[str, FieldType] = {}
        if evaluation_context:
            if evaluation_context.targeting_key:
                eval_context["targeting_key"] = evaluation_context.targeting_key
            # add other fields to eval_context from evaluationContext
            for key, value in evaluation_context.attributes.items():
                eval_context[key] = value
        return self.confidence_sdk.with_context(eval_context)



import dataclasses


@dataclasses.dataclass
class FlagName(object):
    flag: str

    @classmethod
    def parse(cls, resource_name: str) -> "FlagName":
        components = resource_name.split("/", 2)
        if components[0] != "flags":
            raise ValueError("name error")
        return cls(components[1])

    def __str__(self) -> str:
        return f"flags/{self.flag}"


@dataclasses.dataclass
class VariantName(object):
    flag: str
    variant: str

    @classmethod
    def parse(cls, resource_name: str) -> "VariantName":
        components = resource_name.split("/")
        if components[0] != "flags" or components[2] != "variants":
            raise ValueError("name error")
        return cls(components[1], components[3])



import requests_mock
import unittest
import json

from openfeature.flag_evaluation import Reason

import confidence.confidence
from confidence.confidence import Confidence
from confidence.openfeature_provider import ConfidenceOpenFeatureProvider
from confidence.openfeature_provider import EvaluationContext
from confidence.openfeature_provider import Region


class TestMyProvider(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.provider = ConfidenceOpenFeatureProvider(Confidence(client_secret="test"))

    def test_region_has_endpoint(self):
        assert Region.GLOBAL.endpoint()

    def test_resolve_string_with_dot_notation(self):
        ctx = EvaluationContext(targeting_key="boop")
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
                evaluation_context=ctx,
            )

            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.string-key"
            )
            self.assertEqual(result.value, "outer-string")
            self.assertEqual(result.variant, "enabled")
            self.assertEqual(result.reason, Reason.TARGETING_MATCH)

    def test_resolve_failed(self):
        ctx = EvaluationContext(targeting_key="boop")
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=NO_MATCH_STRING_FLAG_RESOLVE,
            )
            result = self.provider.resolve_string_details(
                flag_key="some-flag-that-doesnt-exist",
                default_value="yellow",
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.DEFAULT)
            self.assertEqual(
                result.flag_metadata["flag_key"], "some-flag-that-doesnt-exist"
            )
            self.assertEqual(result.value, "yellow")
            self.assertIsNone(result.variant)

    def test_resolve_string_with_dot_notation_request_payload(self):
        ctx = EvaluationContext(
            targeting_key="boop",
            attributes={"user": {"country": "US"}, "connection": "wifi"},
        )
        with requests_mock.Mocker() as mock:
            confidence.confidence.__version__ = "v0.0.0"
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            self.provider.resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="yellow",
                evaluation_context=ctx,
            )

            last_request = mock.request_history[-1]
            self.assertEqual(last_request.method, "POST")
            self.assertEqual(
                last_request.url, "https://resolver.confidence.dev/v1/flags:resolve"
            )
            self.assertEqual(
                last_request.json(),
                EXPECTED_REQUEST_PAYLOAD,
            )

    def test_resolve_successful_custom_url(self):
        self.provider = ConfidenceOpenFeatureProvider(
            Confidence(
                client_secret="test", custom_resolve_base_url="https://custom_url"
            )
        )
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://custom_url/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_object_details(
                flag_key="python-flag-1.struct-key",
                default_value={"key": "value"},
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.struct-key"
            )
            self.assertEqual(result.value, {"string-key": "inner-string"})

    def test_resolve_response_object_details(self):
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_object_details(
                flag_key="python-flag-1",
                default_value={"key": "value"},
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1")
            self.assertEqual(
                result.value,
                {
                    "double-key": 42.42,
                    "enabled": True,
                    "int-key": 42,
                    "string-key": "outer-string",
                    "struct-key": {"string-key": "inner-string"},
                },
            )

    def test_resolve_struct_details(self):
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_object_details(
                flag_key="python-flag-1.struct-key",
                default_value={"key": "value"},
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.struct-key"
            )
            self.assertEqual(result.value, {"string-key": "inner-string"})

    def test_resolve_integer_details(self):
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_integer_details(
                flag_key="python-flag-1.int-key",
                default_value=-1,
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.int-key")
            self.assertEqual(result.value, 42)

    def test_resolve_boolean_details(self):
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_boolean_details(
                flag_key="python-flag-1.enabled",
                default_value=False,
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(result.flag_metadata["flag_key"], "python-flag-1.enabled")
            self.assertEqual(result.value, True)

    def test_resolve_float_details(self):
        ctx = EvaluationContext(
            targeting_key="boop",
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_float_details(
                flag_key="python-flag-1.double-key",
                default_value=0.01,
                evaluation_context=ctx,
            )

            self.assertEqual(result.reason, Reason.TARGETING_MATCH)
            self.assertEqual(
                result.flag_metadata["flag_key"], "python-flag-1.double-key"
            )
            self.assertEqual(result.value, 42.42)

    def test_resolve_without_targeting_key(self):
        ctx = EvaluationContext(
            attributes={"user": {"country": "US"}, "connection": "wifi"}
        )
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=SUCCESSFUL_FLAG_RESOLVE,
            )
            result = self.provider.resolve_string_details(
                flag_key="python-flag-1.string-key",
                default_value="brown",
                evaluation_context=ctx,
            )
            self.assertIsNotNone(result.value)

            last_request = mock.request_history[-1]
            self.assertIsNone(
                last_request.json()["evaluationContext"].get("targeting_key"),
            )
            self.assertEqual(
                last_request.json()["evaluationContext"].get("connection"), "wifi"
            )

    def test_no_segment_match(self):
        ctx = EvaluationContext(attributes={"connection": "wifi"})
        with requests_mock.Mocker() as mock:
            mock.post(
                "https://resolver.confidence.dev/v1/flags:resolve",
                json=NO_SEGMENT_MATCH_STRING_FLAG_RESOLVE,
            )
            result = self.provider.resolve_string_details(
                flag_key="test-flag.color",
                default_value="brown",
                evaluation_context=ctx,
            )
            self.assertEqual(result.value, "brown")
            self.assertEqual(result.reason, Reason.DEFAULT)

    if __name__ == "__main__":
        unittest.main()


NO_MATCH_STRING_FLAG_RESOLVE = json.loads(
    """{"resolvedFlags": [], "resolveToken": ""}"""
)

NO_SEGMENT_MATCH_STRING_FLAG_RESOLVE = json.loads(
    """{
  "resolveToken": "",
  "resolvedFlags": [
    {
      "flag": "flags/test-flag-2",
      "reason": "RESOLVE_REASON_NO_SEGMENT_MATCH",
      "variant": ""
    }
  ]
}"""
)

EXPECTED_REQUEST_PAYLOAD = json.loads(
    """{
  "clientSecret": "test",
  "evaluationContext": {
    "targeting_key": "boop",
    "user": { "country": "US" },
    "connection": "wifi"
  },
  "apply": true,
  "flags": ["flags/python-flag-1"],
  "sdk": {
    "id": "SDK_ID_PYTHON_CONFIDENCE",
    "version": "v0.0.0"
   }
}"""
)

SUCCESSFUL_FLAG_RESOLVE = json.loads(
    """{
     "resolvedFlags": [
    {
      "flag": "flags/python-flag-1",
      "variant": "flags/python-flag-1/variants/enabled",
      "value": {
        "struct-key": {
          "string-key": "inner-string"
        },
        "string-key": "outer-string",
        "double-key": 42.42,
        "int-key": 42,
        "enabled": true
      },
      "flagSchema": {
        "schema": {
          "struct-key": {
            "structSchema": {
              "schema": {
                "string-key": {
                  "stringSchema": {}
                }
              }
            }
          },
          "string-key": {
            "stringSchema": {}
          },
          "double-key": {
            "doubleSchema": {}
          },
          "int-key": {
            "intSchema": {}
          },
          "enabled": {
            "boolSchema": {}
          }
        }
      },
      "reason": "RESOLVE_REASON_MATCH"
    }
  ],
  "resolveToken": ""
    }"""
)



import asyncio
import uuid

from confidence.confidence import Confidence


async def get_flag():
    root = Confidence("API CLIENT", timeout_ms=100)
    random_uuid = uuid.uuid4()
    uuid_string = str(random_uuid)
    confidence = root.with_context({"targeting_key": uuid_string})
    #confidence.with_context({"app": "python"}).track("navigate", {})
    #print("Tracked navigate event")

    details = confidence.resolve_string_details("hawkflag.color", "default")
    print(f"Flag value: {details.value}")
    print(f"Flag reason: {details.reason}")
    print(f"Flag error code: {details.error_code}")
    print(f"Flag error message: {details.error_message}")


# Another asynchronous function that calls the first one
async def main():
    await get_flag()
    print("Finished calling get_flag")
    await asyncio.sleep(1)
    print("Finished sleeping for 1 seconds")


def sync_main():
    print("Running main asynchronously")


# Run the main function using asyncio.run (Python 3.7+)
if __name__ == "__main__":
    asyncio.run(main())
    sync_main()



from __future__ import annotations

import sys
import typing
from dataclasses import dataclass, field

from confidence.errors import ErrorCode

if sys.version_info >= (3, 11):
    # re-export needed for type checking
    from enum import StrEnum as StrEnum  # noqa: PLC0414
else:
    from enum import Enum

    class StrEnum(str, Enum):
        """
        Backport StrEnum for Python <3.11
        """

        pass


class FlagType(StrEnum):
    BOOLEAN = "BOOLEAN"
    STRING = "STRING"
    OBJECT = "OBJECT"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"


class Reason(StrEnum):
    CACHED = "CACHED"
    DEFAULT = "DEFAULT"
    DISABLED = "DISABLED"
    ERROR = "ERROR"
    STATIC = "STATIC"
    SPLIT = "SPLIT"
    TARGETING_MATCH = "TARGETING_MATCH"
    UNKNOWN = "UNKNOWN"
    TIMEOUT = "TIMEOUT"


FlagMetadata = typing.Mapping[str, typing.Any]

T_co = typing.TypeVar("T_co", covariant=True)


@dataclass
class FlagEvaluationDetails(typing.Generic[T_co]):
    flag_key: str
    value: T_co
    variant: typing.Optional[str] = None
    flag_metadata: FlagMetadata = field(default_factory=dict)
    reason: typing.Optional[typing.Union[str, Reason]] = None
    error_code: typing.Optional[ErrorCode] = None
    error_message: typing.Optional[str] = None


U_co = typing.TypeVar("U_co", covariant=True)


@dataclass
class FlagResolutionDetails(typing.Generic[U_co]):
    value: U_co
    error_code: typing.Optional[ErrorCode] = None
    error_message: typing.Optional[str] = None
    reason: typing.Optional[typing.Union[str, Reason]] = None
    variant: typing.Optional[str] = None
    flag_metadata: FlagMetadata = field(default_factory=dict)



import unittest

from confidence.names import FlagName, VariantName


class NamesTest(unittest.TestCase):
    def test_flag_name_valid(self):
        self.assertEqual(str(FlagName("myFlag")), "flags/myFlag")

    def test_flag_parse(self):
        flag_name = FlagName.parse("flags/myFlag")
        self.assertEqual(flag_name.flag, "myFlag")
        self.assertEqual(str(flag_name), "flags/myFlag")

    def test_flag_parse_error(self):
        self.assertRaises(ValueError, FlagName.parse, "myFlag")

    def test_variant_parse(self):
        variant = VariantName.parse("flags/test-flag-2/variants/variant-1")
        self.assertEqual(variant.flag, "test-flag-2")
        self.assertEqual(variant.variant, "variant-1")

    def test_variant_parse_error(self):
        self.assertRaises(ValueError, VariantName.parse, "myFlag/variants/variant-1")
        self.assertRaises(ValueError, VariantName.parse, "flag/test-flag-2/variant-1")
        self.assertRaises(ValueError, VariantName.parse, "variant-1")


if __name__ == "__main__":
    unittest.main()



import typing
from enum import Enum


class ErrorCode(Enum):
    NOT_READY = "NOT_READY"
    FLAG_NOT_FOUND = "FLAG_NOT_FOUND"
    PARSE_ERROR = "PARSE_ERROR"
    TYPE_MISMATCH = "TYPE_MISMATCH"
    TARGETING_KEY_MISSING = "TARGETING_KEY_MISSING"
    INVALID_CONTEXT = "INVALID_CONTEXT"
    GENERAL = "GENERAL"
    TIMEOUT = "TIMEOUT"


class ConfidenceError(Exception):
    """
    A generic confidence exception, this exception should not be raised. Instead
    the more specific exceptions extending this one should be used.
    """

    def __init__(
        self, error_code: ErrorCode, error_message: typing.Optional[str] = None
    ):
        """
        Constructor for the generic ConfidenceError.
        @param error_message: an optional string message representing why the
        error has been raised
        @param error_code: the ErrorCode string enum value for the type of error
        """
        self.error_message = error_message
        self.error_code = error_code


class TimeoutError(ConfidenceError):
    """
    This exception should be raised when the request to the backend takes longer
    than the timeout set in the client.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the TimeoutError. The error code for this type of exception
        is ErrorCode.TIMEOUT.
        """
        super().__init__(ErrorCode.TIMEOUT, error_message)


class FlagNotFoundError(ConfidenceError):
    """
    This exception should be raised when the SDK cannot find a flag with the
    key provided by the user.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the FlagNotFoundError.  The error code for
        this type of exception is ErrorCode.FLAG_NOT_FOUND.
        @param error_message: an optional string message representing
        why the error has been raised
        """
        super().__init__(ErrorCode.FLAG_NOT_FOUND, error_message)


class GeneralError(ConfidenceError):
    """
    This exception should be raised when the for an exception within the open
    feature python sdk.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the GeneralError.  The error code for this type of exception
        is ErrorCode.GENERAL.
        @param error_message: an optional string message representing why the error
        has been raised
        """
        super().__init__(ErrorCode.GENERAL, error_message)


class ParseError(ConfidenceError):
    """
    This exception should be raised when the flag returned by the provider cannot
    be parsed into a FlagEvaluationDetails object.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the ParseError. The error code for this type of exception
        is ErrorCode.PARSE_ERROR.
        @param error_message: an optional string message representing why the
        error has been raised
        """
        super().__init__(ErrorCode.PARSE_ERROR, error_message)


class TypeMismatchError(ConfidenceError):
    """
    This exception should be raised when the flag returned by the provider does
    not match the type requested by the user.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the TypeMismatchError. The error code for this type of
        exception is ErrorCode.TYPE_MISMATCH.
        @param error_message: an optional string message representing why the
        error has been raised
        """
        super().__init__(ErrorCode.TYPE_MISMATCH, error_message)


class TargetingKeyMissingError(ConfidenceError):
    """
    This exception should be raised when the provider requires a targeting key
    but one was not provided in the evaluation context.
    """

    def __init__(self, error_message: typing.Optional[str] = None):
        """
        Constructor for the TargetingKeyMissingError. The error code for this type of
        exception is ErrorCode.TARGETING_KEY_MISSING.
        @param error_message: a string message representing why the error has been
        raised
        """
        super().__init__(ErrorCode.TARGETING_KEY_MISSING, error_message)


class InvalidContextError(ConfidenceError):
    """
    This exception should be raised when the evaluation context does not meet provider
    requirements.
    """

    def __init__(self, error_message: typing.Optional[str]):
        """
        Constructor for the InvalidContextError. The error code for this type of
        exception is ErrorCode.INVALID_CONTEXT.
        @param error_message: a string message representing why the error has been
        raised
        """
        super().__init__(ErrorCode.INVALID_CONTEXT, error_message)
